@@ -1,6 +1,9 @@
 SELECT setting::integer < 90600 AS pre_96
 FROM pg_settings WHERE name = 'server_version_num';
 
+/* Run tests as unprivileged user */
+SET ROLE TO periods_unprivileged_user;
+
 /* Ensure tables with periods are persistent */
 CREATE UNLOGGED TABLE log (id bigint, s date, e date);
 SELECT periods.add_period('log', 'p', 's', 'e'); -- fails
